"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.LinesAndColumns = exports.Context = exports.ScriptsSourceCode = void 0;
const fs_1 = __importDefault(require("fs"));
const path_1 = __importDefault(require("path"));
const script_let_1 = require("./script-let");
const let_directive_collection_1 = require("./let-directive-collection");
const resolve_parser_1 = require("../parser/resolve-parser");
class ScriptsSourceCode {
    constructor(script, attrs) {
        this._vcode = null;
        this.raw = script;
        this.attrs = attrs;
        this.separateSemiIndex = script.length;
    }
    get vcode() {
        if (this._vcode == null) {
            return this.raw;
        }
        return this._vcode;
    }
    addLet(letCode) {
        if (this._vcode == null) {
            this._vcode = this.raw.trimEnd();
            this.separateSemiIndex = this._vcode.length;
            this._vcode += ";";
            const after = this.raw.slice(this._vcode.length);
            this._vcode += after;
        }
        const start = this._vcode.length;
        this._vcode += letCode;
        return {
            start,
            end: this._vcode.length,
        };
    }
}
exports.ScriptsSourceCode = ScriptsSourceCode;
class Context {
    constructor(code, parserOptions) {
        this.tokens = [];
        this.comments = [];
        this.locsMap = new Map();
        this.letDirCollections = new let_directive_collection_1.LetDirectiveCollections();
        this.state = {};
        this.code = code;
        this.parserOptions = parserOptions;
        this.locs = new LinesAndColumns(code);
        let templateCode = "";
        let scriptCode = "";
        let scriptAttrs = {};
        let start = 0;
        for (const block of extractBlocks(code)) {
            const before = code.slice(start, block.codeRange[0]);
            const blankCode = block.code.replace(/[^\n\r ]/g, " ");
            templateCode += before + blankCode;
            if (block.tag === "script") {
                scriptCode += before.replace(/[^\n\r ]/g, " ") + block.code;
                scriptAttrs = Object.assign(scriptAttrs, block.attrs);
            }
            else {
                scriptCode += before.replace(/[^\n\r ]/g, " ") + blankCode;
            }
            start = block.codeRange[1];
        }
        const after = code.slice(start);
        templateCode += after;
        scriptCode += after.replace(/[^\n\r ]/g, " ");
        this.sourceCode = {
            template: templateCode,
            scripts: new ScriptsSourceCode(scriptCode, scriptAttrs),
        };
        this.scriptLet = new script_let_1.ScriptLetContext(this);
    }
    getLocFromIndex(index) {
        let loc = this.locsMap.get(index);
        if (!loc) {
            loc = this.locs.getLocFromIndex(index);
            this.locsMap.set(index, loc);
        }
        return {
            line: loc.line,
            column: loc.column,
        };
    }
    /**
     * Get the location information of the given node.
     * @param node The node.
     */
    getConvertLocation(node) {
        const { start, end } = node;
        return {
            range: [start, end],
            loc: {
                start: this.getLocFromIndex(start),
                end: this.getLocFromIndex(end),
            },
        };
    }
    addComment(comment) {
        this.comments.push(comment);
    }
    /**
     * Add token to tokens
     */
    addToken(type, range) {
        const token = Object.assign({ type, value: this.getText(range) }, this.getConvertLocation(range));
        this.tokens.push(token);
        return token;
    }
    /**
     * get text
     */
    getText(range) {
        return this.code.slice(range.start, range.end);
    }
    isTypeScript() {
        var _a, _b;
        if (this.state.isTypeScript != null) {
            return this.state.isTypeScript;
        }
        const lang = this.sourceCode.scripts.attrs.lang;
        if (!lang) {
            return (this.state.isTypeScript = false);
        }
        const parserName = resolve_parser_1.getParserName(this.sourceCode.scripts.attrs, (_a = this.parserOptions) === null || _a === void 0 ? void 0 : _a.parser);
        if (parserName === "@typescript-eslint/parser") {
            return (this.state.isTypeScript = true);
        }
        if (parserName.includes("@typescript-eslint/parser")) {
            let targetPath = parserName;
            while (targetPath) {
                const pkgPath = path_1.default.join(targetPath, "package.json");
                if (fs_1.default.existsSync(pkgPath)) {
                    try {
                        return (this.state.isTypeScript =
                            ((_b = JSON.parse(fs_1.default.readFileSync(pkgPath, "utf-8"))) === null || _b === void 0 ? void 0 : _b.name) === "@typescript-eslint/parser");
                    }
                    catch (_c) {
                        return (this.state.isTypeScript = false);
                    }
                }
                const parent = path_1.default.dirname(targetPath);
                if (targetPath === parent) {
                    break;
                }
                targetPath = parent;
            }
        }
        return (this.state.isTypeScript = false);
    }
}
exports.Context = Context;
/** Extract <script> blocks */
function* extractBlocks(code) {
    var _a;
    const startTagRe = /<(script|style)(\s[\s\S]*?)?>/giu;
    const endScriptTagRe = /<\/script(?:\s[\s\S]*?)?>/giu;
    const endStyleTagRe = /<\/style(?:\s[\s\S]*?)?>/giu;
    let startTagRes;
    while ((startTagRes = startTagRe.exec(code))) {
        const [startTag, tag, attributes = ""] = startTagRes;
        const startTagStart = startTagRes.index;
        const startTagEnd = startTagStart + startTag.length;
        const endTagRe = tag.toLowerCase() === "script" ? endScriptTagRe : endStyleTagRe;
        endTagRe.lastIndex = startTagRe.lastIndex;
        const endTagRes = endTagRe.exec(code);
        if (endTagRes) {
            const endTagStart = endTagRes.index;
            const codeRange = [startTagEnd, endTagStart];
            const attrRe = /(?<key>[^\s=]+)(?:=(?:"(?<val1>[^"]*)"|'(?<val2>[^"]*)'|(?<val3>[^\s=]+)))?/giu;
            const attrs = {};
            let attrRes;
            while ((attrRes = attrRe.exec(attributes))) {
                attrs[attrRes.groups.key] =
                    (_a = (attrRes.groups.val1 ||
                        attrRes.groups.val2 ||
                        attrRes.groups.val3)) !== null && _a !== void 0 ? _a : null;
            }
            yield {
                code: code.slice(...codeRange),
                codeRange,
                attrs,
                tag: tag,
            };
            startTagRe.lastIndex = endTagRe.lastIndex;
        }
    }
}
class LinesAndColumns {
    constructor(code) {
        const len = code.length;
        const lineStartIndices = [0];
        for (let index = 0; index < len; index++) {
            const c = code[index];
            if (c === "\r") {
                const next = code[index + 1] || "";
                if (next === "\n") {
                    index++;
                }
                lineStartIndices.push(index + 1);
            }
            else if (c === "\n") {
                lineStartIndices.push(index + 1);
            }
        }
        this.lineStartIndices = lineStartIndices;
    }
    getLocFromIndex(index) {
        const lineNumber = sortedLastIndex(this.lineStartIndices, index);
        return {
            line: lineNumber,
            column: index - this.lineStartIndices[lineNumber - 1],
        };
    }
    getIndexFromLoc(loc) {
        const lineStartIndex = this.lineStartIndices[loc.line - 1];
        const positionIndex = lineStartIndex + loc.column;
        return positionIndex;
    }
}
exports.LinesAndColumns = LinesAndColumns;
/**
 * Uses a binary search to determine the highest index at which value should be inserted into array in order to maintain its sort order.
 */
function sortedLastIndex(array, value) {
    let lower = 0;
    let upper = array.length;
    while (lower < upper) {
        const mid = Math.floor(lower + (upper - lower) / 2);
        const target = array[mid];
        if (target < value) {
            lower = mid + 1;
        }
        else if (target > value) {
            upper = mid;
        }
        else {
            return mid + 1;
        }
    }
    return upper;
}
